
package parser;

import java.util.*;
import parser.*;
import parser.SimpleScriptSyntax.*;

aspect SimpleScriptAspect {
    int Root.run() {
        // 1つ目の関数を呼び出す
        Function func = (Function) entities().get(0);
        return func.call(new int[0]);
    }
    
    // 名前と曖昧性を解決する
    void Root.fix() {
        fix(Collections.EMPTY_MAP, null);
    }
    
    // 再帰的に fix メソッドを呼び出す
    void Node.fix(Map nameToEntity, Node parent) {
        fixChildren(nameToEntity, parent);
    }
    void Node.fixChildren(Map nameToEntity, Node parent) {
        Iterator it = getChildNodes().iterator();
        while (it.hasNext()) {
            Node child = (Node) it.next();
            child.fix(nameToEntity, this);
        }
    }
    // ただし、Root や Function の場合は、関数名やパラメタ名を記号表へ追加する
    void NamedEntityContainer.fix(Map nameToEntity, Node parent) {
        nameToEntity = new HashMap(nameToEntity);
        Iterator it = entities().iterator();
        while (it.hasNext()) {
            NamedEntity entity = (NamedEntity) it.next();
            nameToEntity.put(entity.name().getImage(), entity);
        }
        super.fix(nameToEntity, parent);
    }
    // また、ParameterUse や FunctionCall の場合、fixName を呼び出す
    void NameReferer.fix(Map nameToEntity, Node parent) {
        fixName(nameToEntity, parent);
        super.fix(nameToEntity, parent);
    }
    
    // 名前を解決する
    abstract void NameReferer.fixName(Map nameToEntity, Node parent);
    Function FunctionCall.function;
    void FunctionCall.fixName(Map nameToEntity, Node parent) {
        Object entity = nameToEntity.get(name().getImage());
        if (entity instanceof Function) {
            // 名前は関数をさしている→変数に覚える
            this.function = (Function) entity;
        } else {
            // 名前は関数をさしていない→例外を起こす
            throw new NoSuchElementException(name().toString());
        }
    }
    void ParameterUse.fixName(Map nameToEntity, Node parent) {
        Object entity = nameToEntity.get(name().getImage());
        if (entity instanceof Parameter) {
            // 名前はパラメタをさしている
            // 覚える必要はないので、なにもしない
        } else {
            // 名前はパラメタをさしていない→例外を起こす
            throw new NoSuchElementException(name().toString());
        }
    }
    
    // 曖昧性を解決する
    void ParameterOrCall.fix(Map nameToEntity, Node parent) {
        assert getChildNodes().size() == 2;
        NoSuchElementException lastException = null;
        Iterator it = getChildNodes().iterator();
        while (it.hasNext()) {
            NameReferer child = (NameReferer) it.next();
            // ここで、child は ParameterUse か FunctionCall
            try {
                child.fixName(nameToEntity, parent);
            } catch (NoSuchElementException x) {
                // 名前の解決に失敗した→続行する
                lastException = x;
                continue;
            }
            // 成功した→このabsorberをそのノードで置き換え、サブツリーもfixする
            ((Default.Node) parent).replaceChild(this, child);
            child.fixChildren(nameToEntity, parent);
            return;
        }
        
        // その名前はパラメタ名でも関数名でもない
        throw lastException;
    }
    
    
    // 評価
    
    int Function.call(int[] arguments) {
        Map context = new HashMap();
        for (int i = 0; i < arguments.length; i++) {
            Parameter param = (Parameter) entities().get(i);
            context.put(param.name().getImage(), new Integer(arguments[i]));
        }
        return expression().evaluate(context);
    }
    
    abstract int Expression.evaluate(Map context);
    
    int IfElse.evaluate(Map context) {
        if (condition().evaluate(context) != 0) {
            return thenPart().evaluate(context);
        } else {
            return elsePart().evaluate(context);
        }
    }
    
    int Equal.evaluate(Map context) {
        return operand1().evaluate(context) == operand2().evaluate(context) ? 1 : 0;
    }
    int Unequal.evaluate(Map context) {
        return operand1().evaluate(context) != operand2().evaluate(context) ? 1 : 0;
    }
    int Less.evaluate(Map context) {
        return operand1().evaluate(context) < operand2().evaluate(context) ? 1 : 0;
    }
    int LessEqual.evaluate(Map context) {
        return operand1().evaluate(context) <= operand2().evaluate(context) ? 1 : 0;
    }
    int Greater.evaluate(Map context) {
        return operand1().evaluate(context) > operand2().evaluate(context) ? 1 : 0;
    }
    int GreaterEqual.evaluate(Map context) {
        return operand1().evaluate(context) >= operand2().evaluate(context) ? 1 : 0;
    }
    
    int Addition.evaluate(Map context) {
        return operand1().evaluate(context) + operand2().evaluate(context);
    }
    int Subtraction.evaluate(Map context) {
        return operand1().evaluate(context) - operand2().evaluate(context);
    }
    
    int Multiplication.evaluate(Map context) {
        return operand1().evaluate(context) * operand2().evaluate(context);
    }
    int Division.evaluate(Map context) {
        return operand1().evaluate(context) / operand2().evaluate(context);
    }
    
    int Literal.evaluate(Map context) {
        return Integer.parseInt(value().getImage());
    }
    int ParameterUse.evaluate(Map context) {
        Integer value = (Integer) context.get(name().getImage());
        return value.intValue();
    }
    int FunctionCall.evaluate(Map context) {
        int[] arguments = new int[arguments().size()];
        for (int i = 0; i < arguments.length; i++) {
            arguments[i] = ((Expression) arguments().get(i)).evaluate(context);
        }
        return function.call(arguments);
    }
}

