
/* 
 * $Id: LRTableGenerator.java,v 1.14 2003/12/08 09:56:06 koto Exp $
 * 
 * $Log: LRTableGenerator.java,v $
 * Revision 1.14  2003/12/08 09:56:06  koto
 * Remove embed feature.
 *
 * Revision 1.13  2003/12/08 09:42:52  koto
 * Refactor; call DeterministicState core.
 *
 * Revision 1.12  2003/12/02 14:58:06  koto
 * Remove conflict warning.
 *
 * Revision 1.11  2003/12/02 14:27:19  koto
 * Remove $ambiguous and $absorber.
 *
 * Revision 1.10  2003/09/04 05:30:11  koto
 * Make ambiguity be represented by symbol rather than reduction/LRState.
 *
 * Revision 1.9  2003/09/04 03:54:55  koto
 * Remove absorber LRState/Reduction
 *
 * Revision 1.8  2003/09/04 03:42:44  koto
 * Disuse List list(List).
 *
 * Revision 1.7  2003/08/21 06:05:27  koto
 * Conflict check is turned off if the syntax is decleared as $ambiguous.
 *
 * Revision 1.6  2003/08/20 12:57:22  koto
 * Create absorbers for ambiguous aliases.
 *
 * Revision 1.5  2003/08/20 09:38:47  koto
 * Update to process multiple shift.
 *
 * Revision 1.4  2003/08/17 17:02:20  koto
 * Group links on CFG that have the same labels but different symbols.
 * 
 */

package jp.gr.java_conf.koto.notavacc.lrg;

import java.util.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;
import jp.gr.java_conf.koto.notavacc.dfa.*;
import jp.gr.java_conf.koto.notavacc.parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Original.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.Root;

public final class LRTableGenerator extends Module {
    public LRTableGenerator(Environment env) {
        super(env);
    }
    
    private Root root;
    private DeterministicFiniteAutomaton dfa;
    public LRTable generate(Root root, DeterministicFiniteAutomaton dfa) {
        this.root = root;
        this.dfa = dfa;
        
        /* 
         * Notes:
         * 
         * The output of the module is a table of the LR(1) parsing   
         * algorithm.  An LRTable is the table from states, which are 
         * represented by LRState, and symbols, represented by Symbol,
         * to actions.
         * 
         * Generally, an LR(1) state has LR(1) items and an item is a 
         * triplet of a CFG production, a position in the left-hand   
         * side of the production, and look-a-head symbols.  An LR(1) 
         * item without look-a-head symbols is called a core.
         * 
         * In my LRTableGenerator, a production is a pair of a left-  
         * hand-side nonterminal symbol, and a DFA that represent the 
         * right-hand-side reguler expression on the terminal and     
         * nonterminal symbols.  We call a state of the DFA a core.   
         * And an item has not only look-a-head symbols but also      
         * temporal/additional information.
         * 
         * The nonterminalToInitialCore indicates the map from left-  
         * hand-side nonterminals to initial state of the right-hand- 
         * side DFA.  Initially, the productions are of the extended  
         * BNF, and they are converted into unextended BNF (i.e. CFG) 
         * by buildCFG().  And then, we build the result table from   
         * the unextended-style productions using LRStateBuilder,     
         * LookaheadDeliverer, ReductionGenerator, etc.
         * 
         */
        
        logger.verbose(environment.processingFile, "generating an extended CFG.");
        Map nonterminalToInitialCore = new LinkedHashMap();
        {
            NondeterministicState2 finalState = new NondeterministicState2();
            finalState.isFinal = true;
            
            Iterator it = root.typeDefinitions().iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                if (def.abstractKeyword() == null) {
                    Nonterminal nonterminal = getNonterminal(def);
                    NondeterministicState2 initialState = new NondeterministicState2();
                    link(initialState, def.expression(), finalState);
                    DeterministicConverter converter = new DeterministicConverter();
                    DeterministicState initialCore = converter.createDeterministicState(Collections.singleton(initialState));
                    //initialCore = buildMinimalForm(initialCore);
                    nonterminalToInitialCore.put(nonterminal, initialCore);
                }
            }
            Iterator it7 = root.aliasDefinitions().iterator();
            while (it7.hasNext()) {
                AliasDefinition def = (AliasDefinition) it7.next();
                TreeStackSet argumentLabels = TreeStackSet.EMPTY_SET;
                TreeStackSet labels = TreeStackSet.EMPTY_SET;
                if (root.hasExplicitParameterLabel(def)) {
                    argumentLabels = TreeStackSet.push(argumentLabels, "$label");
                } else {
                    labels = TreeStackSet.push(labels, "$label");
                }
                
                Nonterminal nonterminal = getNonterminal(def);
                NondeterministicState2 initialState = new NondeterministicState2();
                link(argumentLabels, labels, initialState, def.expression(), finalState);
                DeterministicConverter converter = new DeterministicConverter();
                DeterministicState initialCore = converter.createDeterministicState(Collections.singleton(initialState));
                //initialCore = buildMinimalForm(initialCore);
                nonterminalToInitialCore.put(nonterminal, initialCore);
            }
        }
        if (logger.hasError())
            logger.fatal();
        Set allCores = getAllCoreSet(nonterminalToInitialCore);
        nonterminalToInitialCore = buildMinimalForm(nonterminalToInitialCore, allCores);
        allCores = getAllCoreSet(nonterminalToInitialCore);
        logger.verbose(environment.processingFile, "generated.  {0} productions, {1} states.", new Integer(nonterminalToInitialCore.size()), new Integer(allCores.size()));
        if (logger.hasError())
            logger.fatal();
        //reportCores(nonterminalToInitialCore, allCores);
        
        logger.verbose(environment.processingFile, "checking dead symbols.");
        rejectDeadSymbols(nonterminalToInitialCore);
        logger.verbose(environment.processingFile, "checked.");
        if (logger.hasError())
            logger.fatal();
        
        /*
        logger.verbose(environment.processingFile, "checking ambiguity.");
        checkAmbiguity(nonterminalToInitialCore);
        logger.verbose(environment.processingFile, "checked.");
        if (logger.hasError())
            logger.fatal();
        */
        
        logger.verbose(environment.processingFile, "generating a CFG.");
        nonterminalToInitialCore = buildCFG(nonterminalToInitialCore, allCores);
        allCores = getAllCoreSet(nonterminalToInitialCore);
        //logger.verbose(environment.processingFile, "{0} productions.", new Integer(nonterminalToInitialCore.size()));
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(allCores.size()));
        reportCores(nonterminalToInitialCore, allCores);
        
        logger.verbose(environment.processingFile, "checking ambiguity.");
        checkAmbiguity(nonterminalToInitialCore);
        logger.verbose(environment.processingFile, "checked.");
        if (logger.hasError())
            logger.fatal();
        
        Map initialCoreToNonterminal = new HashMap(nonterminalToInitialCore.size());
        Iterator it25 = nonterminalToInitialCore.entrySet().iterator();
        while (it25.hasNext()) {
            Map.Entry entry = (Map.Entry) it25.next();
            initialCoreToNonterminal.put(entry.getValue(), entry.getKey());
        }
        assert initialCoreToNonterminal.size() == nonterminalToInitialCore.size();
        
        logger.verbose(environment.processingFile, "generating LR states.");
        DeterministicState extendedFinalCore = new DeterministicState();
        allCores.add(extendedFinalCore);
        LRHelper lrHelper;
        {
            Map parsableTypeNonterminalToInitialState = new LinkedHashMap();
            LRStateBuilder builder = new LRStateBuilder(nonterminalToInitialCore);
            Set parsableTypeDefinitions = new LinkedHashSet();
            Iterator it = root.typeDefinitions().iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                if (def.parsableKeyword() != null)
                    parsableTypeDefinitions.add(def);
            }
            
            it = parsableTypeDefinitions.iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                TypeNonterminal nonterminal = getNonterminal(def);
                //DeterministicState initialCore = (DeterministicState) nonterminalToInitialCore.get(nonterminal);
                
                DeterministicState extendedInitialCore = new DeterministicState();
                DeterministicState extendedMediumCore = new DeterministicState();
                
                extendedInitialCore.link(new Link(Collections.EMPTY_SET, nonterminal), extendedMediumCore);
                allCores.add(extendedInitialCore);
                
                extendedMediumCore.link(new Link(Collections.EMPTY_SET, Terminal.EOF), extendedFinalCore);
                allCores.add(extendedMediumCore);
                
                LRState state = builder.createLRState(Collections.singleton(extendedInitialCore));
                parsableTypeNonterminalToInitialState.put(nonterminal, state);
            }
            LRState finalState = builder.createLRState(Collections.singleton(extendedFinalCore));
            lrHelper = new LRHelper(parsableTypeNonterminalToInitialState, finalState);
        }
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(lrHelper.getAllLRStateSet().size()));
        //reportLRStates(lrHelper);
        
        logger.verbose(environment.processingFile, "delivering lookahead terminals.");
        {
            LookaheadDeliverer deliverer = new LookaheadDeliverer(nonterminalToInitialCore, allCores);
            deliverer.deliverLookahead(lrHelper);
        }
        logger.verbose(environment.processingFile, "delivered.");
        
        {
            logger.verbose(environment.processingFile, "generating reductions.");
            ReductionGenerator generator = new ReductionGenerator(nonterminalToInitialCore, initialCoreToNonterminal, allCores);
            Set conflictedStates = generator.generateReductions(lrHelper);
            logger.verbose(environment.processingFile, "generated.  {0} LALR(1)-conflicted state(s).", new Integer(conflictedStates.size()));
            reportLRStates(lrHelper);
        }
        
        logger.verbose(environment.processingFile, "generating LR table.");
        LRTable result = buildLRTable(lrHelper);
        logger.verbose(environment.processingFile, "generated.");
        
        return result;
    }
    
    
    private Map nonterminalToHint = new LinkedHashMap();
    private Token getHint(Nonterminal nonterminal) {
        return (Token) nonterminalToHint.get(nonterminal);
    }
    private Map aliasDefinitionToNonterminal = new LinkedHashMap();
    private AliasNonterminal getNonterminal(Parser.AliasDefinition definition) {
        AliasNonterminal result = (AliasNonterminal) aliasDefinitionToNonterminal.get(definition);
        if (result == null) {
            result = new AliasNonterminal(root.getCanonicalName(definition));
            aliasDefinitionToNonterminal.put(definition, result);
            nonterminalToHint.put(result, definition.identifier());
        }
        return result;
    }
    private Map typeDefinitionToNonterminal = new LinkedHashMap();
    private TypeNonterminal getNonterminal(Parser.TypeDefinition definition) {
        TypeNonterminal result = (TypeNonterminal) typeDefinitionToNonterminal.get(definition);
        if (result == null) {
            result = new TypeNonterminal(root.getCanonicalName(definition));
            typeDefinitionToNonterminal.put(definition, result);
            nonterminalToHint.put(result, definition.identifier());
        }
        return result;
    }
    
    
    
    private static class Link {
        public Set labels;
        public Symbol symbol;
        public Link(Set labels, Symbol symbol) {
            this.labels = labels;
            this.symbol = symbol;
        }
        
        public int hashCode() {
            return labels.hashCode() ^ symbol.hashCode();
        }
        public boolean equals(Object rhs) {
            if (rhs == null)
                return false;
            if (rhs instanceof Link)
                equals((Link) rhs);
            return false;
        }
        public boolean equals(Link rhs) {
            return labels.equals(rhs.labels) && symbol.equals(rhs.symbol);
        }
        
        public String toString() {
            StringBuffer result = new StringBuffer();
            Iterator it = labels.iterator();
            while (it.hasNext()) {
                String label = (String) it.next();
                result.append(label);
                result.append(':');
            }
            result.append(symbol);
            return result.toString();
        }
    }
    
    private static class DeterministicState {   // called 'core' rather than state.
        public boolean isFinal = false;
        public Object getBoundAttributes() {
            return new Boolean(isFinal);
        }
        public void copyBoundAttributesFrom(DeterministicState s) {
            isFinal = s.isFinal;
        }
        
        public Map linkToNext = new LinkedHashMap();
        public void link(Link link, DeterministicState state) {
            assert !linkToNext.containsKey(link);
            linkToNext.put(link, state);
        }
        
        private static int numberOfInstance = 0;
        public final int idNumber = numberOfInstance++;
        public String toString() {
            return "State" + idNumber;
        }
        
        // support epsilonMoveStates to make it easy to switch a DeterministicState into a nondeterministic state.
        public Set epsilonMoveStates = null;
        public void epsilonLink(DeterministicState state) {
            if (epsilonMoveStates == null)
                epsilonMoveStates = new LinkedHashSet();
            epsilonMoveStates.add(state);
        }
    }
    private static class NondeterministicState2 extends DeterministicState {
    }
    private Set getAllCoreSet(DeterministicState initialCore) {
        Set result = new LinkedHashSet();
        getAllCoreSet(result, Collections.singleton(initialCore));
        return result;
    }
    private Set getAllCoreSet(Map nonterminalToInitialCore) {
        Set result = new LinkedHashSet();
        getAllCoreSet(result, nonterminalToInitialCore.values());
        return result;
    }
    private void getAllCoreSet(Set result, Collection cores) {
        Iterator it4 = cores.iterator();
        while (it4.hasNext()) {
            DeterministicState core = (DeterministicState) it4.next();
            if (result.add(core)) {
                getAllCoreSet(result, core.linkToNext.values());
                if (core.epsilonMoveStates != null)
                    getAllCoreSet(result, core.epsilonMoveStates);
            }
        }
    }
    
    private void reportCores(Map nonterminalToInitialCore, Set allCores) {
        if (logger.verboseLevel() < Logger.LEVEL_VERBOSEST)
            return;
        
        logger.verbosest(environment.processingFile, "----Low States Report----");
        Iterator it = nonterminalToInitialCore.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            DeterministicState core = (DeterministicState) entry.getValue();
            logger.verbosest("{0} => {1}", nonterminal, core);
        }
        it = allCores.iterator();
        while (it.hasNext()) {
            DeterministicState core = (DeterministicState) it.next();
            logger.verbosest("-- {0} --", core);
            
            if (core.isFinal)
                logger.verbosest("    FINAL STATE");
            Iterator it2 = core.linkToNext.entrySet().iterator();
            while (it2.hasNext()) {
                Map.Entry entry = (Map.Entry) it2.next();
                Link link = (Link) entry.getKey();
                DeterministicState nextCore = (DeterministicState) entry.getValue();
                logger.verbosest("        {0} ==> {1}", link, nextCore);
            }
            if (core.epsilonMoveStates != null && !core.epsilonMoveStates.isEmpty()) {
                logger.verbosest("        epsilon");
                Iterator it51 = core.epsilonMoveStates.iterator();
                while (it51.hasNext()) {
                    DeterministicState nextCore = (DeterministicState) it51.next();
                    logger.verbosest("            ==> {0}", nextCore);
                }
            }
        }
    }
    
    private void link(NondeterministicState2 initialState, Expression expression, NondeterministicState2 finalState) {
        link(TreeStackSet.EMPTY_SET, TreeStackSet.EMPTY_SET, initialState, expression, finalState);
    }
    private void link(TreeStackSet argumentLabels, TreeStackSet labels, NondeterministicState2 initialState, Expression expression, NondeterministicState2 finalState) {
        ExpressionVisitor.Context initialContext = new ExpressionVisitor.Context();
        initialContext.labelStrings = labels;
        initialContext.argumentLabelStrings = argumentLabels;
        new ExpressionVisitor(root, 0) {
            protected void visit(Context context, SelectiveExpression expr, Object argument) {
                NondeterministicState2[] connectionPoints = (NondeterministicState2[]) argument;
                NondeterministicState2 initialState = connectionPoints[0];
                NondeterministicState2 finalState = connectionPoints[1];
                
                Iterator it = expr.operands().iterator();
                while (it.hasNext()) {
                    Expression operand = (Expression) it.next();
                    NondeterministicState2 tmp1 = new NondeterministicState2();
                    NondeterministicState2 tmp2 = new NondeterministicState2();
                    
                    initialState.epsilonLink(tmp1);
                    visitAll(context, operand, new NondeterministicState2[] { tmp1, tmp2 });
                    tmp2.epsilonLink(finalState);
                }
            }
            protected void visit(Context context, SequentialExpression expr, Object argument) {
                NondeterministicState2[] connectionPoints = (NondeterministicState2[]) argument;
                NondeterministicState2 initialState = connectionPoints[0];
                NondeterministicState2 finalState = connectionPoints[1];
                
                if (expr.operands().isEmpty()) {
                    initialState.epsilonLink(finalState);
                } else {
                    NondeterministicState2 head = initialState;
                    Iterator it = expr.operands().iterator();
                    while (it.hasNext()) {
                        Expression operand = (Expression) it.next();
                        NondeterministicState2 tail;
                        if (it.hasNext())
                            tail = new NondeterministicState2();
                        else
                            tail = finalState;
                        visitAll(context, operand, new NondeterministicState2[] { head, tail });
                        head = tail;
                    }
                }
            }
            protected void visit(Context context, PlusExpression expr, Object argument) {
                NondeterministicState2[] connectionPoints = (NondeterministicState2[]) argument;
                NondeterministicState2 initialState = connectionPoints[0];
                NondeterministicState2 finalState = connectionPoints[1];
                
                NondeterministicState2 tmp1 = new NondeterministicState2();
                NondeterministicState2 tmp2 = new NondeterministicState2();
                initialState.epsilonLink(tmp1);
                visitAll(context, expr.operand(), new NondeterministicState2[] { tmp1, tmp2 });
                tmp2.epsilonLink(finalState);
                tmp2.epsilonLink(tmp1);
            }
            protected void visit(Context context, IdentifierExpression expr, AliasDefinition def, Object argument) {
                link(context, getNonterminal(def), argument);
            }
            protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
                if (def.abstractKeyword() != null)
                    logger.error("The type {0} is abstract.", expr.name().identifier());
                link(context, getNonterminal(def), argument);
            }
            protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
                link(context, new NamedTerminal(root.getCanonicalName(def)), argument);
            }
            protected void visit(Context context, StringExpression expr, Object argument) {
                link(context, dfa.getTerminal(expr), argument);
            }
            private void link(Context context, Symbol symbol, Object argument) {
                NondeterministicState2[] connectionPoints = (NondeterministicState2[]) argument;
                NondeterministicState2 initialState = connectionPoints[0];
                NondeterministicState2 finalState = connectionPoints[1];
                
                Link link = new Link(context.labelStrings, symbol);
                initialState.link(link, finalState);
            }
        }.visitAll(initialContext, expression, new NondeterministicState2[] { initialState, finalState });
    }
    
    /*private void link(TreeStackSet argumentLabels, TreeStackSet labels, NondeterministicState initialState, Expression expression, NondeterministicState finalState) {
        // if argumentLabels == null, 'expression' is in AbsorberExpression, so cannot use labels.
        
        if (expression instanceof SelectiveExpression) {
            SelectiveExpression exp = (SelectiveExpression) expression;
            Iterator it = exp.operands().iterator();
            while (it.hasNext()) {
                Expression operand = (Expression) it.next();
                link(argumentLabels, labels, initialState, operand, finalState);
            }
        } else if (expression instanceof SequentialExpression) {
            SequentialExpression exp = (SequentialExpression) expression;
            if (exp.operands().isEmpty()) {
                initialState.epsilonLink(finalState);
            } else {
                NondeterministicState cursor = initialState;
                Iterator it = exp.operands().iterator();
                while (it.hasNext()) {
                    Expression operand = (Expression) it.next();
                    NondeterministicState tmp;
                    if (it.hasNext())
                        tmp = new NondeterministicState();
                    else
                        tmp = finalState;
                    link(argumentLabels, labels, cursor, operand, tmp);
                    cursor = tmp;
                }
            }
        } else if (expression instanceof PlusExpression) {
            PlusExpression exp = (PlusExpression) expression;
            NondeterministicState tmp1 = new NondeterministicState();
            NondeterministicState tmp2 = new NondeterministicState();
            initialState.epsilonLink(tmp1);
            link(argumentLabels, labels, tmp1, exp.operand(), tmp2);
            tmp2.epsilonLink(finalState);
            tmp2.epsilonLink(tmp1);
        } else if (expression instanceof RestrictorExpression) {
            RestrictorExpression exp = (RestrictorExpression) expression;
            link(argumentLabels, labels, initialState, exp.operand(), finalState);
        } else if (expression instanceof LabeledExpression) {
            LabeledExpression exp = (LabeledExpression) expression;
            if (argumentLabels == null) {
                logger.error("{0}; cannot use any label in $absorber expressions.", exp.label());
            } else {
                if (exp.label().getImage().equals("$label"))
                    labels = TreeStackSet.pushAll(labels, argumentLabels);
                else
                    labels = TreeStackSet.push(labels, exp.label().getImage());
                link(argumentLabels, labels, initialState, exp.operand(), finalState);
            }
        } else if (expression instanceof AbsorberExpression) {
            AbsorberExpression exp = (AbsorberExpression) expression;
            Link link = new Link(labels, getNonterminal(exp));
            initialState.link(link, finalState);
        } else if (expression instanceof StringExpression) {
            StringExpression exp = (StringExpression) expression;
            Terminal terminal = dfa.getTerminal(exp);
            Link link = new Link(labels, terminal);
            initialState.link(link, finalState);
        } else {
            IdentifierExpression exp = (IdentifierExpression) expression;
            link(argumentLabels, labels, initialState, exp, finalState);
        }
    }
    private void link(TreeStackSet argumentLabels, TreeStackSet labels, NondeterministicState initialState, IdentifierExpression exp, NondeterministicState finalState) {
        Definition def = root.getDefinition(exp.name());
        if (def instanceof TypeDefinition) {
            TypeDefinition d = (TypeDefinition) def;
            if (d.abstractKeyword() != null)
                logger.error("The type {0} is abstract.", exp.name().identifier());
            Link link = new Link(labels, getNonterminal(d));
            initialState.link(link, finalState);
        } else if (def instanceof AliasDefinition) {
            AliasDefinition d = (AliasDefinition) def;
            Link link = new Link(labels, getNonterminal(d));
            initialState.link(link, finalState);
        } else {
            TokenReservation d = (TokenReservation) def;
            Link link = new Link(labels, new NamedTerminal(root.getCanonicalName(d)));
            initialState.link(link, finalState);
        }
    }*/
    
    
    private class DeterministicConverter {
        private Map substatesToCore = new LinkedHashMap();
        public DeterministicState createDeterministicState(Set substates) {
            substates = getEpsilonClosure(substates);
            DeterministicState result = (DeterministicState) substatesToCore.get(substates);
            if (result == null) {
                result = new DeterministicState();
                substatesToCore.put(substates, result);
                bind(substates, result);
            }
            return result;
        }
        
        private void bind(Set substates, DeterministicState target) {
            // set final
            Iterator it49 = substates.iterator();
            while (it49.hasNext()) {
                DeterministicState substate = (DeterministicState) it49.next();
                if (substate.isFinal) {
                    target.isFinal = true;
                    break;
                }
            }
            
            // set map
            MultiMap pairToLinks = new MultiMap();
            MultiMap pairToSubstates = new MultiMap();
            Iterator it29 = substates.iterator();
            while (it29.hasNext()) {
                DeterministicState substate = (DeterministicState) it29.next();
                Iterator it3 = substate.linkToNext.entrySet().iterator();
                while (it3.hasNext()) {
                    Map.Entry entry = (Map.Entry) it3.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState state = (DeterministicState) entry.getValue();
                    List pair = Arrays.asList(new Object[] { link.labels, link.symbol });
                    pairToLinks.add(pair, link);
                    pairToSubstates.add(pair, state);
                }
            }
            
            MultiMap linkToSubstates = new MultiMap();
            Iterator it30 = pairToLinks.entrySet().iterator();
            while (it30.hasNext()) {
                Map.Entry entry = (Map.Entry) it30.next();
                List pair = (List) entry.getKey();
                Set links = (Set) entry.getValue();
                Set states = pairToSubstates.getValueSet(pair);
                
                DeterministicState newCore = createDeterministicState(states);
                target.link((Link) links.iterator().next(), newCore);
            }
        }
        
        private Set getEpsilonClosure(Set substates) {
            Set result = new LinkedHashSet();
            getEpsilonClosure(result, substates);
            return result;
        }
        private void getEpsilonClosure(Set result, Set substates) {
            Iterator it1 = substates.iterator();
            while (it1.hasNext()) {
                DeterministicState substate = (DeterministicState) it1.next();
                if (result.add(substate) && substate.epsilonMoveStates != null)
                    getEpsilonClosure(result, substate.epsilonMoveStates);
            }
        }
    }
    
    //private DeterministicState buildMinimalForm(DeterministicState initialCore) {
    //    Map nonterminalToInitialCore = Collections.singletonMap(null, initialCore);
    //    Set allCores = getAllCoreSet(nonterminalToInitialCore);
    //    nonterminalToInitialCore = buildMinimalForm(nonterminalToInitialCore, allCores);
    //    return (DeterministicState) nonterminalToInitialCore.get(null);
    //}
    private Map buildMinimalForm(Map nonterminalToInitialCore, Set allCores) {
        Set initialCores = new HashSet(nonterminalToInitialCore.values());
        DeterministicState deadCore = new DeterministicState();
        
        // initialize
        List groups = new ArrayList(allCores.size() + 1);
        Map coreToGroupID = new HashMap();
        {
            Map attributeToGroupID = new LinkedHashMap();
            
            attributeToGroupID.put(deadCore.getBoundAttributes(), new Integer(0));
            groups.add(new LinkedList(Collections.singleton(deadCore)));
            coreToGroupID.put(deadCore, new Integer(0));
            
            Iterator it = allCores.iterator();
            while (it.hasNext()) {
                DeterministicState core = (DeterministicState) it.next();
                assert core.epsilonMoveStates == null;
                Object key = core.getBoundAttributes();
                Integer groupID;
                if (initialCores.contains(core)) {
                    groupID = new Integer(groups.size());
                    groups.add(new LinkedList());
                } else {
                    groupID = (Integer) attributeToGroupID.get(key);
                    if (groupID == null) {
                        groupID = new Integer(groups.size());
                        groups.add(new LinkedList());
                        attributeToGroupID.put(key, groupID);
                    }
                }
                List cores = (List) groups.get(groupID.intValue());
                cores.add(core);
                coreToGroupID.put(core, groupID);
            }
        }
        
        // loop
        int lastMoodifiedIndex = -1;
        int index = 0;
        for (;; index++) {
            if (index >= groups.size()) {
                if (lastMoodifiedIndex == -1)
                    break;
                index -= groups.size();
            }
            
            List group = (List) groups.get(index);
            assert !group.isEmpty();
            if (group.size() <= 1) {
                if (index == lastMoodifiedIndex)
                    break;
                continue;
            }
            
            Map keyToCoreList = new LinkedHashMap();
            Iterator it = group.iterator();
            while (it.hasNext()) {
                DeterministicState core = (DeterministicState) it.next();
                Set key = new HashSet();
                Iterator it2 = core.linkToNext.entrySet().iterator();
                while (it2.hasNext()) {
                    Map.Entry entry = (Map.Entry) it2.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState next = (DeterministicState) entry.getValue();
                    key.add(Arrays.asList(new Object[] { link.symbol, link.labels, coreToGroupID.get(next) }));
                }
                List list = (List) keyToCoreList.get(key);
                if (list == null) {
                    list = new LinkedList();
                    keyToCoreList.put(key, list);
                }
                list.add(core);
            }
            
            if (keyToCoreList.size() <= 1) {
                if (index == lastMoodifiedIndex)
                    break;
                continue;
            }
            
            lastMoodifiedIndex = index;
            it = keyToCoreList.values().iterator();
            while (it.hasNext()) {
                List cores = (List) it.next();
                if (it.hasNext()) {
                    Integer newID = new Integer(groups.size());
                    groups.add(cores);
                    Iterator it2 = cores.iterator();
                    while (it2.hasNext()) {
                        DeterministicState core = (DeterministicState) it2.next();
                        coreToGroupID.put(core, newID);
                    }
                } else {
                    groups.set(index, cores);
                }
            }
        }
        
        // create result states.
        Integer deadIndexInteger = (Integer) coreToGroupID.get(deadCore);
        int deadIndex = deadIndexInteger.intValue();
        DeterministicState[] result = new DeterministicState[groups.size()];
        for (int i = 0; i < result.length; i++) {
            if (i != deadIndex)
                result[i] = new DeterministicState();
        }
        
        for (int i = 0; i < result.length; i++) {
            if (i != deadIndex) {
                List group = (List) groups.get(i);
                DeterministicState hot = result[i];
                
                DeterministicState old = (DeterministicState) group.get(0);
                hot.copyBoundAttributesFrom(old);
                
                Iterator it46 = old.linkToNext.entrySet().iterator();
                while (it46.hasNext()) {
                    Map.Entry entry = (Map.Entry) it46.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState next = (DeterministicState) entry.getValue();
                    Integer nextGroupID = (Integer) coreToGroupID.get(next);
                    if (nextGroupID.intValue() != deadIndex) {
                        DeterministicState hotNext = result[nextGroupID.intValue()];
                        hot.link(link, hotNext);
                    }
                }
            }
        }
        
        // create result map.
        Map resultMap = new LinkedHashMap();
        Iterator it = nonterminalToInitialCore.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            DeterministicState core = (DeterministicState) entry.getValue();
            Integer groupID = (Integer) coreToGroupID.get(core);
            resultMap.put(nonterminal, result[groupID.intValue()]);
        }
        return resultMap;
    }
    
    private void rejectDeadSymbols(Map nonterminalToInitialCore) {
        List nonterminals = new LinkedList(nonterminalToInitialCore.keySet());
        Set liveNonterminals = new HashSet();
        for (;;) {
            boolean progress = false;
            
            Iterator it = nonterminals.iterator();
            while (it.hasNext()) {
                Nonterminal nonterminal = (Nonterminal) it.next();
                if (isLiveNonterminal(nonterminal, (DeterministicState) nonterminalToInitialCore.get(nonterminal), liveNonterminals)) {
                    liveNonterminals.add(nonterminal);
                    it.remove();
                    progress = true;
                }
            }
            
            if (!progress)
                break;
        }
        if (!nonterminals.isEmpty()) {
            Iterator it91 = nonterminals.iterator();
            while (it91.hasNext()) {
                Nonterminal nonterminal = (Nonterminal) it91.next();
                logger.error("{0} is a dead symbol; it cannot match any input.", getHint(nonterminal));
            }
        }
    }
    private boolean isLiveNonterminal(Nonterminal nonterminal, DeterministicState initialCore, Set liveNonterminals) {
        Set done = new HashSet();
        LinkedList undone = new LinkedList();
        undone.add(initialCore);
        while (!undone.isEmpty()) {
            DeterministicState core = (DeterministicState) undone.removeFirst();
            if (done.add(core)) {
                if (core.isFinal)
                    return true;
                Iterator it89 = core.linkToNext.entrySet().iterator();
                while (it89.hasNext()) {
                    Map.Entry entry = (Map.Entry) it89.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState next = (DeterministicState) entry.getValue();
                    if (liveNonterminals.contains(link.symbol)) {
                        undone.add(next);
                    } else {
                        if (link.symbol instanceof Terminal) {
                            liveNonterminals.add(link.symbol);
                            undone.add(next);
                        }
                    }
                }
            }
        }
        return false;
    }
    
    private void checkAmbiguity(Map nonterminalToInitialCore) {
        
    }
    
    
    /*
    private Map cloneCores(Map nonterminalToInitialCore) {
        Map result = new LinkedHashMap();
        Iterator it43 = nonterminalToInitialCore.entrySet().iterator();
        while (it43.hasNext()) {
            Map.Entry entry = (Map.Entry) it43.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            DeterministicState initialCore = (DeterministicState) entry.getValue();
            result.put(nonterminal, cloneCores(initialCore));
        }
        return result;
    }
    private DeterministicState cloneCores(DeterministicState initialCore) {
        Map oldToNew = new HashMap();
        Iterator it40 = getAllCoreSet(initialCore).iterator();
        while (it40.hasNext()) {
            DeterministicState core = (DeterministicState) it40.next();
            assert core.epsilonMoveStates == null;
            DeterministicState hot = new DeterministicState();
            hot.copyBoundAttributesFrom(core);
            oldToNew.put(core, hot);
        }
        Iterator it41 = oldToNew.entrySet().iterator();
        while (it41.hasNext()) {
            Map.Entry entry41 = (Map.Entry) it41.next();
            DeterministicState old = (DeterministicState) entry41.getKey();
            DeterministicState hot = (DeterministicState) entry41.getValue();
            Iterator it42 = old.linkToNext.entrySet().iterator();
            while (it42.hasNext()) {
                Map.Entry entry42 = (Map.Entry) it42.next();
                Link link = (Link) entry42.getKey();
                DeterministicState next = (DeterministicState) entry42.getValue();
                hot.link(link, (DeterministicState) oldToNew.get(next));
            }
        }
        return (DeterministicState) oldToNew.get(initialCore);
    }
    private void checkAmbiguity(Map nonterminalToInitialCore) {
        // remove labels.
        nonterminalToInitialCore = cloneCores(nonterminalToInitialCore);
        Iterator it54 = nonterminalToInitialCore.entrySet().iterator();
        while (it54.hasNext()) {
            Map.Entry entry = (Map.Entry) it54.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            DeterministicState initialCore = (DeterministicState) entry.getValue();
            
            Set allCores = getAllCoreSet(initialCore);
            Iterator it55 = allCores.iterator();
            while (it55.hasNext()) {
                DeterministicState core = (DeterministicState) it55.next();
                Iterator it = core.linkToNext.entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry entry2 = (Map.Entry) it.next();
                    Link link = (Link) entry2.getKey();
                    DeterministicState next = (DeterministicState) entry2.getValue();
                    if (!link.labels.isEmpty()) {
                        it.remove();
                        DeterministicState c = new DeterministicState();
                        core.epsilonLink(c);
                        c.link(new Link(Collections.EMPTY_SET, link.symbol), next);
                    }
                }
            }
            initialCore = new DeterministicConverter().createDeterministicState(Collections.singleton(initialCore));
            initialCore = buildMinimalForm(initialCore);
            entry.setValue(initialCore);
        }
        
        // regularize
        Map tmp = new LinkedHashMap(nonterminalToInitialCore);
        Iterator it53 = nonterminalToInitialCore.entrySet().iterator();
        while (it53.hasNext()) {
            Map.Entry entry = (Map.Entry) it53.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            entry.setValue(regularize(nonterminal, tmp));
        }
        
        Set allCores = getAllCoreSet(nonterminalToInitialCore);
        reportCores(nonterminalToInitialCore, allCores);
    }
    private DeterministicState regularize(Nonterminal nonterminal, Map originalMapNonterminalToInitialCore) {
        Set usedNonterminals = new LinkedHashSet();
        {
            LinkedList undone = new LinkedList();
            undone.add(nonterminal);
            while (!undone.isEmpty()) {
                Nonterminal n = (Nonterminal) undone.removeFirst();
                if (usedNonterminals.add(n)) {
                    Set allCores = getAllCoreSet((DeterministicState) originalMapNonterminalToInitialCore.get(n));
                    Iterator it57 = allCores.iterator();
                    while (it57.hasNext()) {
                        DeterministicState c = (DeterministicState) it57.next();
                        Iterator it60 = c.linkToNext.keySet().iterator();
                        while (it60.hasNext()) {
                            Link link = (Link) it60.next();
                            if (link.symbol instanceof Nonterminal)
                                undone.add((Nonterminal) link.symbol);
                        }
                    }
                }
            }
        }
        
        
        // clone and mark.
        final int START = 0;
        final int END = 1;
        Map nonterminalToInitialCore = new LinkedHashMap();
        Map nonterminalToMatterCores = new LinkedHashMap();
        Map nonterminalToAllCores = new LinkedHashMap();
        Iterator it61 = usedNonterminals.iterator();
        while (it61.hasNext()) {
            Nonterminal n = (Nonterminal) it61.next();
            DeterministicState start = (DeterministicState) originalMapNonterminalToInitialCore.get(n);
            start = cloneCores(start);
            nonterminalToInitialCore.put(n, start);
            DeterministicState end = new DeterministicState();
            nonterminalToMatterCores.put(n, new DeterministicState[] { start, end });
            
            Set allCores = getAllCoreSet(start);
            nonterminalToAllCores.put(n, allCores);
            Iterator it55 = allCores.iterator();
            while (it55.hasNext()) {
                DeterministicState core = (DeterministicState) it55.next();
                if (core.isFinal) {
                    core.isFinal = false;
                    core.epsilonLink(end);
                }
            }
            
            if (n == nonterminal) {
                DeterministicState realEnd = new DeterministicState();
                realEnd.isFinal = true;
                end.epsilonLink(realEnd);
            }
        }
        
        // connect.
        Iterator it62 = usedNonterminals.iterator();
        while (it62.hasNext()) {
            Nonterminal n = (Nonterminal) it62.next();
            DeterministicState start = (DeterministicState) nonterminalToInitialCore.get(n);
            
            Set allCores = getAllCoreSet(start);
            Iterator it57 = allCores.iterator();
            while (it57.hasNext()) {
                DeterministicState c = (DeterministicState) it57.next();
                Iterator it = c.linkToNext.entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry entry = (Map.Entry) it.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState next = (DeterministicState) entry.getValue();
                    if (link.symbol instanceof Nonterminal) {
                        DeterministicState[] matterCores = (DeterministicState[]) nonterminalToMatterCores.get((Nonterminal) link.symbol);
                        
                        it.remove();
                        c.epsilonLink(matterCores[START]);
                        matterCores[END].epsilonLink(next);
                    }
                }
            }
        }
        
        DeterministicState result = (DeterministicState) nonterminalToInitialCore.get(nonterminal);
        result = new DeterministicConverter().createDeterministicState(Collections.singleton(result));
        result = buildMinimalForm(result);
        return result;
    }*/
    /*private DeterministicState regularize(DeterministicState initialCore, Set regularNonterminals, Map nonterminalToInitialCore) {
        initialCore = cloneCores(initialCore);
        
        Set allCores = getAllCoreSet(initialCore);
        Iterator it45 = allCores.iterator();
        while (it45.hasNext()) {
            DeterministicState core = (DeterministicState) it45.next();
            Iterator it = core.linkToNext.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry) it.next();
                Link link = (Link) entry.getKey();
                DeterministicState next = (DeterministicState) entry.getValue();
                if (link.symbol instanceof Nonterminal) {
                    Nonterminal nonterminal = (Nonterminal) link.symbol;
                    if (!regularNonterminals.contains(nonterminal))
                        return null;
                    it.remove();
                    
                    DeterministicState subInitial = (DeterministicState) nonterminalToInitialCore.get(nonterminal);
                    subInitial = cloneCores(subInitial);
                    Set subAll = getAllCoreSet(subInitial);
                    
                    core.epsilonLink(subInitial);
                    Iterator it52 = subAll.iterator();
                    while (it52.hasNext()) {
                        DeterministicState sub = (DeterministicState) it52.next();
                        if (sub.isFinal)
                            sub.epsilonLink(next);
                    }
                }
            }
            
        }
        
        initialCore = new DeterministicConverter().createDeterministicState(Collections.singleton(initialCore));
        initialCore = buildMinimalForm(initialCore);
        return initialCore;
    }*/
    
    
    private Map coreToMapPreviousCoreToLinks(Set allCores) {
        Map result = new LinkedHashMap();
        Iterator it13 = allCores.iterator();
        while (it13.hasNext()) {
            DeterministicState core = (DeterministicState) it13.next();
            Iterator it47 = core.linkToNext.entrySet().iterator();
            while (it47.hasNext()) {
                Map.Entry entry = (Map.Entry) it47.next();
                Link link = (Link) entry.getKey();
                DeterministicState next = (DeterministicState) entry.getValue();
                
                MultiMap map = (MultiMap) result.get(next);
                if (map == null) {
                    map = new MultiMap();
                    result.put(next, map);
                }
                map.add(core, link);
            }
        }
        return result;
    }
    
    private Map buildCFG(Map nonterminalToInitialCore, Set allCores) {
        Iterator it95 = allCores.iterator();
        while (it95.hasNext()) {
            DeterministicState core = (DeterministicState) it95.next();
            if (core.linkToNext.isEmpty()) {
                if (!core.isFinal)
                    throw new AssertionError(core + " : " + core.isFinal + " : " + core.linkToNext);
            }
        }
        
        Map result = new LinkedHashMap();
        Iterator it10 = nonterminalToInitialCore.entrySet().iterator();
        while (it10.hasNext()) {
            Map.Entry entry = (Map.Entry) it10.next();
            Nonterminal nonterminal = (Nonterminal) entry.getKey();
            DeterministicState initialCore = (DeterministicState) entry.getValue();
            result.putAll(new CFGBuilder(nonterminal, initialCore).getMapNonterminalToInitialCore());
        }
        return result;
    }
    private static class CFGBuilder {
        private static final Set LABEL_INHERITOR = Collections.singleton("$label");
        
        private final Map nonterminalToInitialCore = new LinkedHashMap();
        private Map getMapNonterminalToInitialCore() {
            return nonterminalToInitialCore;
        }
        
        public CFGBuilder(Nonterminal nonterminal, DeterministicState initialCore) {
            DeterministicState newCore = duplicate(initialCore, TreeStackSet.EMPTY_SET);
            nonterminalToInitialCore.put(nonterminal, newCore);
        }
        
        private Map coreToAnonymous = new LinkedHashMap();
        private AnonymousNonterminal getAnonymous(DeterministicState startCore) {
            AnonymousNonterminal result = (AnonymousNonterminal) coreToAnonymous.get(startCore);
            if (result != null)
                return result;
            
            result = new AnonymousNonterminal();
            coreToAnonymous.put(startCore, result);
            DeterministicState newCore = duplicate(startCore, TreeStackSet.EMPTY_SET);
            newCore.isFinal = false;
            nonterminalToInitialCore.put(result, newCore);
            return result;
        }
        private DeterministicState duplicate(DeterministicState core, TreeStackSet duplicatedCores) {
            DeterministicState result  = new DeterministicState();
            result.copyBoundAttributesFrom(core);
            if (duplicatedCores.contains(core)) {
                DeterministicState finalCore  = new DeterministicState();
                finalCore.isFinal = true;
                
                result.linkToNext.put(new Link(LABEL_INHERITOR, getAnonymous(core)), finalCore);
            } else {
                duplicatedCores = TreeStackSet.push(duplicatedCores, core);
                
                Iterator it56 = core.linkToNext.entrySet().iterator();
                while (it56.hasNext()) {
                    Map.Entry entry = (Map.Entry) it56.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState next = (DeterministicState) entry.getValue();
                    
                    result.linkToNext.put(link, duplicate(next, duplicatedCores));
                }
            }
            return result;
        }
    }
    
    private static class LRState implements LRTable.LRState {
        public MultiMap symbolToNextStates = new MultiMap();
        public Map symbolToNextStates() {
            return symbolToNextStates;
        }
        public MultiMap symbolToReductions = null;
        public Map symbolToReductions() {
            return symbolToReductions;
        }
        
        public Set kernelItems = new LinkedHashSet();
        public Set items = new LinkedHashSet();
        public Map coreToItem = new LinkedHashMap();
        public boolean containsFinal = false;
        public void discardTemporalValues() {
            kernelItems = null;
            items = null;
            coreToItem = null;
        }
        
        private static int numberOfInstance = 0;
        public final int idNumber = numberOfInstance++;
        public String toString() {
            return "LRState" + idNumber;
        }
    }
    private static class Item {
        public final LRState owner;
        public final DeterministicState core;
        public final boolean kernel;
        
        public Set lookaheads = null;   // The look-a-head of the production.  If not core.isFinal, this is the same as the nextItem's one.
        public MultiMap symbolToNextItems = new MultiMap();
        public Set deliveredItems = new LinkedHashSet();
        
        public Item(LRState owner, DeterministicState core, boolean kernel) {
            assert owner != null && core != null;
            this.owner = owner;
            this.core = core;
            this.kernel = kernel;
        }
        
        public String toString() {
            String k = kernel ? " (kernel)" : "";
            String f = core.isFinal ? " (FINAL)" : "";
            return core + f + " ; " + lookaheads + k;
        }
    }
    
    public static class Reduction implements LRTable.Reduction {
        private Nonterminal reducedSymbol;
        private List labelSetList;
        public Reduction(Nonterminal reducedSymbol, List labelSetList) {
            this.reducedSymbol = reducedSymbol;
            this.labelSetList = labelSetList;
        }
        public Nonterminal reducedSymbol() {
            return reducedSymbol;
        }
        public List labelSetList() {
            return labelSetList;
        }
        
        public int hashCode() {
            return reducedSymbol.hashCode() + labelSetList.hashCode();
        }
        public final boolean equals(Object rhs) {
            if (rhs instanceof Reduction)
                return equals((Reduction) rhs);
            return false;
        }
        public boolean equals(Reduction rhs) {
            return reducedSymbol.equals(rhs.reducedSymbol) && labelSetList.equals(rhs.labelSetList);
        }
        
        public String toString() {
            return reducedSymbol + " : " + labelSetList;
        }
    }
    
    private void reportLRStates(LRHelper lrHelper) {
        if (logger.verboseLevel() < Logger.LEVEL_VERBOSEST)
            return;
        
        logger.verbosest(environment.processingFile, "----High States Report----");
        
        Iterator it = lrHelper.parsableTypeNonterminalToInitialState().entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            TypeNonterminal parsableTypeNonterminal = (TypeNonterminal) entry.getKey();
            LRState state = (LRState) entry.getValue();
            logger.verbosest("{0} => {1}", parsableTypeNonterminal, state);
        }
        
        it = lrHelper.getAllLRStateSet().iterator();
        while (it.hasNext()) {
            LRState state = (LRState) it.next();
            logger.verbosest("-- {0} --", state);
            Iterator it2 = state.symbolToNextStates.entrySet().iterator();
            while (it2.hasNext()) {
                Map.Entry entry = (Map.Entry) it2.next();
                Symbol symbol = (Symbol) entry.getKey();
                Collection nextStates = (Collection) entry.getValue();
                Iterator it27 = nextStates.iterator();
                while (it27.hasNext()) {
                    LRState nextState = (LRState) it27.next();
                    logger.verbosest("        {0} ==> {1}", symbol, nextState);
                }
            }
            logger.verbosest("    -- items --");
            Iterator it34 = state.items.iterator();
            while (it34.hasNext()) {
                Item item = (Item) it34.next();
                logger.verbosest("        {0}", item);
                
                if (item.core.isFinal)
                    logger.verbosest("            FINAL STATE");
                Iterator it35 = item.core.linkToNext.entrySet().iterator();
                while (it35.hasNext()) {
                    Map.Entry entry = (Map.Entry) it35.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState nextCore = (DeterministicState) entry.getValue();
                    logger.verbosest("            {0} ==> {1}", link, nextCore);
                }
                
            }
            if (state.symbolToReductions != null) {
                logger.verbosest("    -- reductions --");
                Iterator it14 = state.symbolToReductions.entrySet().iterator();
                while (it14.hasNext()) {
                    Map.Entry entry = (Map.Entry) it14.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    Set reductions = (Set) entry.getValue();
                    logger.verbosest("        {0} ==>", symbol);
                    Iterator it15 = reductions.iterator();
                    while (it15.hasNext()) {
                        Reduction reduction = (Reduction) it15.next();
                        logger.verbosest("            {0}", reduction);
                    }
                }
            }
        }
    }
    
    private class LRStateBuilder {
        private Map nonterminalToInitialCore;
        public LRStateBuilder(Map nonterminalToInitialCore) {
            this.nonterminalToInitialCore = nonterminalToInitialCore;
        }
        
        private Map kernelCoresToLRState = new HashMap();
        private LRState createLRState(Set kernelCores) {
            LRState result = (LRState) kernelCoresToLRState.get(kernelCores);
            if (result != null)
                return result;  // result.coreToItem shuld be set. (see below)
            
            // create and register.
            result = new LRState();
            kernelCoresToLRState.put(kernelCores, result);
            
            // create kernel items.
            Iterator it12 = kernelCores.iterator();
            while (it12.hasNext()) {
                DeterministicState kernelCore = (DeterministicState) it12.next();
                Item kernel = new Item(result, kernelCore, true);
                result.kernelItems.add(kernel);
                result.coreToItem.put(kernelCore, kernel);
            }
            
            // create non-kernel items.
            LinkedList undone = new LinkedList(result.kernelItems);
            while (!undone.isEmpty()) {
                Item item = (Item) undone.removeFirst();
                result.items.add(item);
                if (item.core.isFinal)
                    result.containsFinal = true;
                Iterator it11 = item.core.linkToNext.keySet().iterator();
                while (it11.hasNext()) {
                    Link link = (Link) it11.next();
                    if (link.symbol instanceof Nonterminal) {
                        DeterministicState core = (DeterministicState) nonterminalToInitialCore.get(link.symbol);
                        Item deliveredItem = (Item) result.coreToItem.get(core);
                        if (deliveredItem == null) {
                            deliveredItem = new Item(result, core, false);
                            result.coreToItem.put(core, deliveredItem);
                            undone.add(deliveredItem);
                        }
                        item.deliveredItems.add(deliveredItem);
                    }
                }
            }
            
            // list the next kernelCores.
            MultiMap symbolToNextKernelCores = new MultiMap();
            Iterator it22 = result.items.iterator();
            while (it22.hasNext()) {
                Item item = (Item) it22.next();
                Iterator it = item.core.linkToNext.entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry entry = (Map.Entry) it.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState nextKernelCore = (DeterministicState) entry.getValue();
                    symbolToNextKernelCores.add(link.symbol, nextKernelCore);
                }
            }
            // create next states.
            Iterator it = symbolToNextKernelCores.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry) it.next();
                Symbol symbol = (Symbol) entry.getKey();
                Set nextKernelCores = (Set) entry.getValue();
                LRState nextState = createLRState(nextKernelCores);     // nextState.coreToItem shuld be set. (see above)
                result.symbolToNextStates.add(symbol, nextState);
            }
            // link items.
            Iterator it16 = result.items.iterator();
            while (it16.hasNext()) {
                Item item = (Item) it16.next();
                Iterator it48 = item.core.linkToNext.entrySet().iterator();
                while (it48.hasNext()) {
                    Map.Entry entry = (Map.Entry) it48.next();
                    Link link = (Link) entry.getKey();
                    DeterministicState nextCore = (DeterministicState) entry.getValue();
                    
                    Collection nextStates = result.symbolToNextStates.getValueSet(link.symbol);
                    assert nextStates.size() == 1;
                    LRState nextState = (LRState) nextStates.iterator().next();
                    Item nextItem = (Item) nextState.coreToItem.get(nextCore);  // nextState.coreToItem shuld be set. (see above)
                    item.symbolToNextItems.add(link.symbol, nextItem);
                }
            }
            
            return result;
        }
    } // class LRStateBuilder
    
    private class LRHelper {
        private Map parsableTypeNonterminalToInitialState;
        private LRState finalState;
        public LRHelper(Map parsableTypeNonterminalToInitialState, LRState finalState) {
            this.parsableTypeNonterminalToInitialState = parsableTypeNonterminalToInitialState;
            this.finalState = finalState;
        }
        
        public Map parsableTypeNonterminalToInitialState() {
            return parsableTypeNonterminalToInitialState;
        }
        public LRState finalState() {
            return finalState;
        }
        
        private Map backsteps = null;
        public Map backstepStateToSymbol(LRState state) {
            getAllLRStateSet();
            Map result = (Map) backsteps.get(state);
            if (result == null) {
                result = Collections.EMPTY_MAP;
                backsteps.put(state, result);
            }
            return result;
        }
        public Set backstep(LRState state) {
            return backstepStateToSymbol(state).keySet();
        }
        
        private Set allLRStates = null;
        public Set getAllLRStateSet() {
            if (allLRStates == null) {
                backsteps = new LinkedHashMap();
                allLRStates = new LinkedHashSet();
                allLRStates.add(finalState);
                LinkedList undone = new LinkedList(parsableTypeNonterminalToInitialState.values());
                while (!undone.isEmpty()) {
                    LRState state = (LRState) undone.removeFirst();
                    if (allLRStates.add(state)) {
                        Iterator it59 = state.symbolToNextStates.entrySet().iterator();
                        while (it59.hasNext()) {
                            Map.Entry entry = (Map.Entry) it59.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            Collection nextStates = (Collection) entry.getValue();
                            Iterator it27 = nextStates.iterator();
                            while (it27.hasNext()) {
                                LRState nextState = (LRState) it27.next();
                                Map map = (Map) backsteps.get(nextState);
                                if (map == null) {
                                    map = new LinkedHashMap();
                                    backsteps.put(nextState, map);
                                }
                                map.put(state, symbol);
                            }
                            undone.addAll(nextStates);
                        }
                    }
                }
            }
            return allLRStates;
        }
        
        /*private Map initialStateToParsableTypeNonterminal = null;
        public Map initialStateToParsableTypeNonterminal() {
            if (initialStateToParsableTypeNonterminal == null) {
                initialStateToParsableTypeNonterminal = new LinkedHashMap();
                Iterator it44 = parsableTypeNonterminalToInitialState.entrySet().iterator();
                while (it44.hasNext()) {
                    Map.Entry entry = (Map.Entry) it44.next();
                    initialStateToParsableTypeNonterminal.put(entry.getValue(), entry.getKey());
                }
                assert initialStateToParsableTypeNonterminal.size() == parsableTypeNonterminalToInitialState.size();
            }
            return initialStateToParsableTypeNonterminal;
        }*/
        
        private MultiMap itemToGeneratorKernels = null;
        public MultiMap itemToGeneratorKernels() {
            if (itemToGeneratorKernels == null) {
                itemToGeneratorKernels = new MultiMap();
                Iterator it12 = getAllLRStateSet().iterator();
                while (it12.hasNext()) {
                    LRState state = (LRState) it12.next();
                    Iterator it31 = state.kernelItems.iterator();
                    while (it31.hasNext()) {
                        Item kernel = (Item) it31.next();
                        Set done = new HashSet();
                        LinkedList undone = new LinkedList();
                        undone.add(kernel);
                        while (!undone.isEmpty()) {
                            Item item = (Item) undone.removeFirst();
                            if (done.add(item)) {
                                Iterator it33 = item.deliveredItems.iterator();
                                while (it33.hasNext()) {
                                    Item deliveredItem = (Item) it33.next();
                                    itemToGeneratorKernels.add(deliveredItem, kernel);
                                    undone.add(deliveredItem);
                                }
                            }
                        }
                    }
                }
            }
            return itemToGeneratorKernels;
        }
        public Set getGeneratorKernels(Item item) {
            return itemToGeneratorKernels().getValueSet(item);
        }
        
        private Map stateToAscendedStateAndSymbol = null;
        public Map stateToAscendedStateAndSymbol() {
            if (stateToAscendedStateAndSymbol == null) {
                stateToAscendedStateAndSymbol = new LinkedHashMap();
                
                LinkedList undone = new LinkedList();
                Iterator it39 = parsableTypeNonterminalToInitialState.values().iterator();
                while (it39.hasNext()) {
                    LRState state = (LRState) it39.next();
                    stateToAscendedStateAndSymbol.put(state, null);
                    undone.add(state);
                }
                
                while (!undone.isEmpty()) {
                    LRState state = (LRState) undone.removeFirst();
                    Iterator it38 = state.symbolToNextStates.entrySet().iterator();
                    while (it38.hasNext()) {
                        Map.Entry entry = (Map.Entry) it38.next();
                        Symbol symbol = (Symbol) entry.getKey();
                        Collection nextStates = (Collection) entry.getValue();
                        Iterator it27 = nextStates.iterator();
                        while (it27.hasNext()) {
                            LRState nextState = (LRState) it27.next();
                            if (!stateToAscendedStateAndSymbol.containsKey(nextState)) {
                                stateToAscendedStateAndSymbol.put(nextState, new Object[] { state, symbol });
                                undone.add(nextState);
                            }
                        }
                    }
                }
            }
            return stateToAscendedStateAndSymbol;
        }
        
    } // class LRHelper
    
    private class LookaheadDeliverer {
        private Map nonterminalToInitialCore;
        private Set allCores;
        public LookaheadDeliverer(Map nonterminalToInitialCore, Set allCores) {
            this.nonterminalToInitialCore = nonterminalToInitialCore;
            this.allCores = allCores;
            buildEmptyAcceptableSet();
            buildFirstSet();
        }
        
        private Set emptyAcceptable;
        private void buildEmptyAcceptableSet() {
            emptyAcceptable = new HashSet();
            Set unfixed = new LinkedHashSet(allCores);
            boolean modified = true;
            while (!unfixed.isEmpty() && modified) {
                modified = false;
                Iterator outer = unfixed.iterator();
                while (outer.hasNext()) {
                    DeterministicState core = (DeterministicState) outer.next();
                    // We try to know whether the core accepts the empty input.
                    
                    int check = -1;      // -1=unmatch the empty input, 0=unfixed, 1=match the empty input.
                    if (core.isFinal) {
                        check = 1;
                    } else {
                        Iterator inner = core.linkToNext.entrySet().iterator();
                        while (inner.hasNext()) {
                            Map.Entry entry = (Map.Entry) inner.next();
                            Link link = (Link) entry.getKey();
                            Symbol symbol = link.symbol;
                            if (symbol instanceof Nonterminal) {
                                DeterministicState next = (DeterministicState) entry.getValue();
                                DeterministicState sub = (DeterministicState) nonterminalToInitialCore.get((Nonterminal) symbol);
                                if (unfixed.contains(next) || unfixed.contains(sub)) {
                                    check = 0;
                                } else {
                                    if (emptyAcceptable.contains(next) && emptyAcceptable.contains(sub)) {
                                        check = 1;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    if (check != 0) {
                        outer.remove();
                        modified = true;
                        if (check == 1) {
                            emptyAcceptable.add(core);
                        }
                    }
                }
            }
        }
        public boolean isEmptyAcceptable(DeterministicState core) {
            return emptyAcceptable.contains(core);
        }
        public boolean isEmptyAcceptable(Nonterminal nonterminal) {
            DeterministicState initialCore = (DeterministicState) nonterminalToInitialCore.get(nonterminal);
            return isEmptyAcceptable(initialCore);
        }
        public boolean isEmptyAcceptable(Symbol symbol) {
            if (symbol instanceof Terminal)
                return false;
            return isEmptyAcceptable((Nonterminal) symbol);
        }
        
        private MultiMap coreToFirstSet;
        private void buildFirstSet() {
            coreToFirstSet = new MultiMap();
            {
                Iterator it0 = allCores.iterator();
                while (it0.hasNext()) {
                    DeterministicState core = (DeterministicState) it0.next();
                    coreToFirstSet.put(core, new LinkedHashSet());
                }
            }
            
            Set unfixed = new LinkedHashSet(allCores);
            boolean modified = true;
            while (!unfixed.isEmpty() && modified) {
                modified = false;
                Iterator outer = unfixed.iterator();
                while (outer.hasNext()) {
                    DeterministicState core = (DeterministicState) outer.next();
                    Set first = coreToFirstSet.getValueSet(core);
                    boolean fixed = true;
                    Iterator inner = core.linkToNext.entrySet().iterator();
                    while (inner.hasNext()) {
                        Map.Entry entry = (Map.Entry) inner.next();
                        Link link = (Link) entry.getKey();
                        if (link.symbol instanceof Terminal) {
                            modified |= first.add(link.symbol);
                        } else {
                            DeterministicState subCore = (DeterministicState) nonterminalToInitialCore.get((Nonterminal) link.symbol);
                            modified |= first.addAll(coreToFirstSet.getValueSet(subCore));
                            if (unfixed.contains(subCore))
                                fixed = false;
                            
                            if (isEmptyAcceptable(subCore)) {
                                DeterministicState nextCore = (DeterministicState) entry.getValue();
                                modified |= first.addAll(coreToFirstSet.getValueSet(nextCore));
                                if (unfixed.contains(nextCore))
                                    fixed = false;
                            }
                        }
                    }
                    if (fixed) {
                        outer.remove();
                        modified = true;
                    }
                }
            }
        }
        public Set first(DeterministicState core) {
            return coreToFirstSet.getValueSet(core);
        }
        
        public void deliverLookahead(LRHelper lrHelper) {
            Iterator it36 = lrHelper.getAllLRStateSet().iterator();
            while (it36.hasNext()) {
                LRState state = (LRState) it36.next();
                Iterator it37 = state.items.iterator();
                while (it37.hasNext()) {
                    Item item = (Item) it37.next();
                    assert item.lookaheads == null;
                    item.lookaheads = new LinkedHashSet();
                }
            }
            
            LinkedList jobs = new LinkedList();
            MultiMap itemToLookaheadDependencies = new MultiMap();
            {
                LinkedList undone = new LinkedList();
                {
                    Iterator it4 = lrHelper.parsableTypeNonterminalToInitialState().values().iterator();
                    while (it4.hasNext()) {
                        LRState state = (LRState) it4.next();
                        undone.addAll(state.kernelItems);
                    }
                }
                Set done = new HashSet();
                while (!undone.isEmpty()) {
                    Item item = (Item) undone.removeFirst();
                    if (done.add(item)) {
                        Iterator it3 = item.symbolToNextItems.entrySet().iterator();
                        while (it3.hasNext()) {
                            Map.Entry entry = (Map.Entry) it3.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            Set nextItems = (Set) entry.getValue();
                            Iterator it50 = nextItems.iterator();
                            while (it50.hasNext()) {
                                Item nextItem = (Item) it50.next();
                                // item --symbol--> nextItem
                                itemToLookaheadDependencies.add(item, nextItem);    // If the item has a look-a-head X, the nextItem should have X.
                                undone.add(nextItem);
                                
                                if (symbol instanceof Nonterminal) {
                                    DeterministicState core = (DeterministicState) nonterminalToInitialCore.get((Nonterminal) symbol);
                                    Item deliveredItem = (Item) item.owner.coreToItem.get(core);
                                    jobs.add(new Object[] { deliveredItem, first(nextItem.core) });
                                    undone.add(deliveredItem);
                                    if (isEmptyAcceptable(nextItem.core))
                                        itemToLookaheadDependencies.add(item, deliveredItem);   // If the item has a look-a-head X, the deliveredItem should have X.
                                }
                            }
                        }
                    }
                }
            }
            
            Iterator it1 = jobs.iterator();
            while (it1.hasNext()) {
                Object[] job = (Object[]) it1.next();
                Item item = (Item) job[0];
                Set lookaheads = (Set) job[1];
                deliverLookahead(itemToLookaheadDependencies, item, lookaheads);
            }
        }
        private void deliverLookahead(MultiMap itemToLookaheadDependencies, Item item, Set lookaheads) {
            if (item.lookaheads.addAll(lookaheads)) {
                Set dependencies = itemToLookaheadDependencies.getValueSet(item);
                if (dependencies != null) {
                    Iterator it6 = dependencies.iterator();
                    while (it6.hasNext()) {
                        Item dependency = (Item) it6.next();
                        deliverLookahead(itemToLookaheadDependencies, dependency, lookaheads);
                    }
                }
            }
        }
    }
    
    private class ReductionGenerator {
        private Map initialCoreToNonterminal;
        private Map coreToMapPreviousCoreToLinks;
        public ReductionGenerator(Map nonterminalToInitialCore, Map initialCoreToNonterminal, Set allCores) {
            this.initialCoreToNonterminal = initialCoreToNonterminal;
            coreToMapPreviousCoreToLinks = coreToMapPreviousCoreToLinks(allCores);
        }
        
        public Set generateReductions(LRHelper lrHelper) {
            Set conflictedStates = new LinkedHashSet();
            Iterator it21 = lrHelper.getAllLRStateSet().iterator();
            while (it21.hasNext()) {
                LRState state = (LRState) it21.next();
                assert state.symbolToReductions == null;
                state.symbolToReductions = new MultiMap();
                boolean conflicted = false;
                if (state.containsFinal) {
                    Iterator it5 = state.items.iterator();
                    while (it5.hasNext()) {
                        Item item = (Item) it5.next();
                        if (item.core.isFinal) {
                            Reduction reduction = getReduction(item);
                            Iterator it24 = item.lookaheads.iterator();
                            while (it24.hasNext()) {
                                Symbol lookahead = (Symbol) it24.next();
                                state.symbolToReductions.add(lookahead, reduction);
                            }
                        }
                    }
                    if (!conflicted) {
                        Set shiftableSymbol = state.symbolToNextStates.keySet();
                        Iterator it26 = state.symbolToReductions.entrySet().iterator();
                        while (it26.hasNext()) {
                            Map.Entry entry = (Map.Entry) it26.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            Set reductions = (Set) entry.getValue();
                            if (shiftableSymbol.contains(symbol) || reductions.size() > 1)
                                conflicted = true;
                        }
                    }
                }
                if (conflicted)
                    conflictedStates.add(state);
            }
            
            return conflictedStates;
        }
        
        private Reduction getReduction(Item item) {
            return getReduction(item.core);
        }
        private Map stateToReduction = new HashMap();
        private Reduction getReduction(DeterministicState finalState) {
            DeterministicState state = finalState;
            Reduction result = (Reduction) stateToReduction.get(finalState);
            if (result == null) {
                LinkedList labelsList = new LinkedList();
                
                for (;;) {
                    Nonterminal nonterminal = (Nonterminal) initialCoreToNonterminal.get(state);
                    if (nonterminal != null) {
                        result = new Reduction(nonterminal, labelsList);
                        break;
                    }
                    
                    MultiMap map = (MultiMap) coreToMapPreviousCoreToLinks.get(state);
                    assert map.size() == 1;
                    state = (DeterministicState) map.keySet().iterator().next();
                    Set links = (Set) map.values().iterator().next();
                    Iterator it = links.iterator();
                    Link link = (Link) it.next();
                    while (it.hasNext()) {
                        Link l = (Link) it.next();
                        assert link.labels.equals(l.labels);
                    }
                    
                    labelsList.addFirst(link.labels);
                }
                
                stateToReduction.put(finalState, result);
            }
            return result;
        }
    }
    
    
    private LRTable buildLRTable(LRHelper lrHelper) {
        return new GeneratedLRTable(lrHelper.parsableTypeNonterminalToInitialState(), lrHelper.finalState());
    }
    private static class GeneratedLRTable extends LRTable {
        private final Map parsableTypeNonterminalToInitialState;
        private final LRState finalState;
        public GeneratedLRTable(Map parsableTypeNonterminalToInitialState, LRState finalState) {
            this.parsableTypeNonterminalToInitialState = Collections.unmodifiableMap(parsableTypeNonterminalToInitialState);
            this.finalState = finalState;
            
            Iterator it28 = states().iterator();
            while (it28.hasNext()) {
                LRTableGenerator.LRState state = (LRTableGenerator.LRState) it28.next();
                state.discardTemporalValues();
            }
        }
        
        public Map symbolToInitialState() {
            return parsableTypeNonterminalToInitialState;
        }
        public LRState finalState() {
            return finalState;
        }
    }
    
}

