
package jp.gr.java_conf.koto.notavacc.parser;

import java.io.*;
import java.util.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;
import jp.gr.java_conf.koto.notavacc.parser.Original.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.Root;

public abstract class ExpressionVisitor {
    private final Root root;
    private final int visitCountLimit;
    protected ExpressionVisitor(Root root) {
        this(root, 0);
    }
    protected ExpressionVisitor(Root root, int visitCountLimit) {
        this.root = root;
        this.visitCountLimit = visitCountLimit;
    }
    
    public static class Context {
        public TreeStack opened;
        public TreeStackSet argumentLabelStrings;
        public TreeStackSet labelStrings;
        public TreeStackSet typeRestrictors;
        public Context() {
            this.opened = TreeStack.EMPTY_STACK;
            this.argumentLabelStrings = TreeStackSet.EMPTY_SET;
            this.labelStrings = TreeStackSet.EMPTY_SET;
            this.typeRestrictors = TreeStackSet.EMPTY_SET;
        }
        public Object clone() {
            Context result = new Context();
            result.opened = opened;
            result.argumentLabelStrings = argumentLabelStrings;
            result.labelStrings = labelStrings;
            result.typeRestrictors = typeRestrictors;
            return result;
        }
    }
    
    public void visitAll(Expression node, Object argument) {
        visitAll(new Context(), node, argument);
    }
    public void visitAll(Context context, Expression node, Object argument) {
        if (node instanceof SelectiveExpression) {
            visit(context, (SelectiveExpression) node, argument);
        } else if (node instanceof SequentialExpression) {
            visit(context, (SequentialExpression) node, argument);
        } else if (node instanceof PlusExpression) {
            visit(context, (PlusExpression) node, argument);
        } else if (node instanceof RestrictorExpression) {
            visit(context, (RestrictorExpression) node, argument);
        } else if (node instanceof LabeledExpression) {
            visit(context, (LabeledExpression) node, argument);
        } else if (node instanceof IdentifierExpression) {
            visit(context, (IdentifierExpression) node, argument);
        } else if (node instanceof StringExpression) {
            visit(context, (StringExpression) node, argument);
        } else {
            throw new AssertionError();
        }
    }
    
    protected void visit(Context context, SelectiveExpression expr, Object argument) {
        Iterator it0 = expr.operands().iterator();
        while (it0.hasNext()) {
            Expression operand = (Expression) it0.next();
            visitAll(context, operand, argument);
        }
    }
    protected void visit(Context context, SequentialExpression expr, Object argument) {
        Iterator it0 = expr.operands().iterator();
        while (it0.hasNext()) {
            Expression operand = (Expression) it0.next();
            visitAll(context, operand, argument);
        }
    }
    protected void visit(Context context, PlusExpression expr, Object argument) {
        visitAll(context, expr.operand(), argument);
    }
    protected void visit(Context context, RestrictorExpression expr, Object argument) {
        Context newContext = (Context) context.clone();
        newContext.typeRestrictors = TreeStackSet.push(context.typeRestrictors, expr.typeRestrictor());
        visitAll(newContext, expr.operand(), argument);
    }
    protected void visit(Context context, LabeledExpression expr, Object argument) {
        Context newContext = (Context) context.clone();
        
        String label = expr.label().getImage();
        if (label.equals("$label")) {
            newContext.labelStrings = TreeStackSet.pushAll(context.labelStrings, context.argumentLabelStrings);
        } else {
            newContext.labelStrings = TreeStackSet.push(context.labelStrings, label);
        }
        
        visitAll(newContext, expr.operand(), argument);
    }
    protected void visit(Context context, IdentifierExpression expr, Object argument) {
        Definition def = root.getDefinition(expr.name());
        if (def instanceof TypeDefinition) {
            visit(context, expr, (TypeDefinition) def, argument);
        } else if (def instanceof AliasDefinition) {
            visit(context, expr, (AliasDefinition) def, argument);
        } else if (def instanceof TokenReservation) {
            visit(context, expr, (TokenReservation) def, argument);
        } else {
            throw new AssertionError();
        }
    }
    protected void visit(Context context, IdentifierExpression expr, AliasDefinition def, Object argument) {
        List key = Arrays.asList(new Object[] { context.labelStrings, def, context.typeRestrictors });
        if (context.opened.count(key) >= visitCountLimit)
            return;
        
        Context newContext = (Context) context.clone();
        newContext.opened = TreeStack.push(context.opened, key);
        if (root.hasExplicitParameterLabel(def)) {
            newContext.argumentLabelStrings = newContext.labelStrings;
            newContext.labelStrings = TreeStackSet.EMPTY_SET;
        } else {
            newContext.argumentLabelStrings = TreeStackSet.EMPTY_SET;
        }
        
        visitAll(newContext, def.expression(), argument);
    }
    protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
        
    }
    protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
        
    }
    protected void visit(Context context, StringExpression expr, Object argument) {
        
    }
    
}

